#include <stdint.h>
#include <stdio.h>

#include "tonc_memdef.h"
#include "interrupt.h"
#include "input.h"
#include "filesystem.h"
#include "image.h"
#include "text.h"
#include "util.h"
#include "sprite.h"
#include "math.h"


#define SHIP_SPEED 8

#define SHIP_ENGINE_PAL_INDEX 5
#define SHIP_ENGINE_OFF_COLOR RGB15(6, 3, 3)
#define SHIP_ENGINE_ON_COLOR RGB15(29, 21, 10)

u32 g_skybox_rot;

VECTOR g_cam_pos;
VECTOR g_pre_rotate_cam_pos;
s32 g_cam_rotation_deg;

s32 g_cam_rotation_sin;
s32 g_cam_rotation_cos;

BG_AFFINE g_bg_affine;

VECTOR g_ship_position;
VECTOR g_ship_velocity;
VECTOR g_heading_z;

VECTOR g_offset;

u8 cross_flag = 0;
u8 first_cross_flag = 1;

u32 g_seconds_left = 7200;

char g_lap_num[] = {"0"};

void update_ship();
void update_camera(); 
void check_for_collisions();
void update_lap_counter();

IWRAM_CODE void vbl_isr() {
	input_poll();
}

#define MODE7_D 160

IWRAM_CODE void hbl_isr() {

	// Mode 7

	// .8 * .16 / .12 = .12
	s32 y_div = (g_cam_pos.y * lut_div(REG_VCOUNT + 1)) >> 12;
	
	// .8 * .12 / .8 = .12
	s32 pa = (g_cam_rotation_cos * y_div) >> 8;
	s32 pc = (g_cam_rotation_sin * y_div) >> 8;

	// .12 / .4 = .8
	g_bg_affine.pa = pa >> 4;
	g_bg_affine.pc = pc >> 4;

	// Horiz
	s32 x = 120 * (pa >> 4);
	s32 y = (MODE7_D * pc) >> 4;
	g_bg_affine.dx = g_cam_pos.x - x + y;
	
	// Vert
	x = 120 * (pc >> 4);
	y = (MODE7_D * pa) >> 4;
	g_bg_affine.dy = g_cam_pos.z - x - y;

	REG_BG_AFFINE[2] = g_bg_affine;

	// Fog

	/*
	u32 ey= g_bg_affine.pb * 6 >> 12;
    if(ey > 16)
        ey = 16;

	REG_BLDALPHA= BLDA_BUILD(16 - ey, ey);
	*/

}

typedef enum {
    NOTE_C = 0, NOTE_CIS, NOTE_D,   NOTE_DIS, 
    NOTE_E,   NOTE_F,   NOTE_FIS, NOTE_G, 
    NOTE_GIS, NOTE_A,   NOTE_BES, NOTE_B
} eSndNoteId;

// Rates for traditional notes in octave +5
const u32 __snd_rates[12]=
{
    8013, 7566, 7144, 6742, // C , C#, D , D#
    6362, 6005, 5666, 5346, // E , F , F#, G
    5048, 4766, 4499, 4246  // G#, A , A#, B
};

#define SND_RATE(note, oct, div) ( 2048 - ((__snd_rates[note] * div) >> (4 + (oct))) )

void sound_sweep(u32 num, u32 dir, u32 time) {

	REG_SND1SWEEP = num | dir | SSW_TIME(time);
	
}

void sound_note(eSndNoteId note, u32 octave, u32 div, u32 vol) {

    // Envelope: vol=12, decay, max step time (7) ; 50% duty
    REG_SND1CNT = SSQR_ENV_BUILD(vol, 0, 0) | SSQR_DUTY3_4;

	// Sample use: note A, octave 0
    REG_SND1FREQ = SFREQ_RESET | SND_RATE(note, octave, div);
	
}

void sound_noise(u32 volume, u32 freq, u32 length) {

	REG_SND4CNT = SSQR_ENV_BUILD(volume, 0, length);
	
	u16 new_freq = SFREQ_RESET | 0b0001 | (freq << 4);
	REG_SND4FREQ = new_freq;

}

void init_sound() {

	REG_SNDSTAT = SSTAT_ENABLE;
	REG_SNDDMGCNT = SDMG_BUILD_LR(SDMG_SQR1 | SDMG_NOISE, 7);
    REG_SNDDSCNT = SDS_DMG100;

	// No sweep
    REG_SND1SWEEP = SSW_OFF;

}

int AgbMain(void) {

	// Interrupt setup
	
	irq_init();

	irq_enable(II_HBLANK, (fnptr)hbl_isr);
	irq_enable(II_VBLANK, (fnptr)vbl_isr);

	init_sound();

	// Filesystem setup

	filesystem_init(FILESYSTEM_SEARCH_ROM);
	filesystem_error_t error = filesystem_get_error();

	// Load a couple tilesets

	filesystem_file_t* fontTilesetFile = filesystem_get_file("font.img");
	error = filesystem_get_error();

	filesystem_file_t* spriteFile = filesystem_get_file("ship.img");
	error = filesystem_get_error();

	filesystem_file_t* tilesetFile = filesystem_get_file("tiles.img");
	error = filesystem_get_error();

	filesystem_file_t* trackFile = filesystem_get_file("track.img");
	error = filesystem_get_error();

	// Load the tileset into charblock 2
    image_tileset_t* fontTileset = image_tileset_from_file(fontTilesetFile);
	image_load_tileset(fontTileset, 2);

	image_tileset_t* spriteTileset = image_tileset_from_file(spriteFile);

	// Load the tileset into charblock 0
	image_tileset_t* tileset = image_tileset_from_file(tilesetFile);
	image_load_tileset(tileset, 0);
	image_load_palette256(tileset, PALETTE_BG);

	// Clear BG0
	for (u32 y = 0; y < 32; ++y) {
		for (u32 x = 0; x < 32; ++x) {
			se_mat[15][y][x] = 0;
		}
	}

	// Clear BG1
	for (u32 y = 0; y < 32; ++y) {
		for (u32 x = 0; x < 32; ++x) {
			se_mat[24][y][x] = 0;
		}
	}

	// Write skybox	
	u32 skybox_start = 32;
	u32 skybox_index = 32;

	for (u32 y = 0; y < 3; ++y) {
		skybox_index = skybox_start + skybox_start * y;

		for (u32 x = 0; x < 32; ++x) {
			se_mat[24][y][x] = skybox_index;
			skybox_index++;
		}		
		
	}


	//lap counter
	char g_seconds_left_str[] = {"100"};
	sprintf(g_seconds_left_str, "%d", g_seconds_left/60);
	text_write("LAP", 15, 0, 0, 0);
	text_write("SECONDS LEFT", 15, 0, 13, 0);
	text_write(g_seconds_left_str, 15, 0, 26, 0);

	//pal_bg_bank[1][1] = RGB15(0, 31, 0);
	
	const u32* track_data = image_get_tilesraw(image_tileset_from_file(trackFile));
	memcpy32(se_mem[25], track_data, 64 * 64 / 4);

	// Mode 0
	REG_DISPCNT = DCNT_MODE1 | DCNT_BG0 | DCNT_BG1 | DCNT_BG2 | DCNT_OAM | DCNT_OBJ_2D;

	// Use BG0, with 4BPP tiles, charblock 2, screenblock 15
	REG_BG0CNT = BG_4BPP | BG_CBB(2) | BG_SBB(15);

	// Use BG1, with 8BPP tiles, charblock 0, screenblock 24
	REG_BG1CNT = BG_8BPP | BG_CBB(0) | BG_SBB(24) | BG_REG_32x32;

	// Use BG2, with 8BPP tiles, charblock 0, screenblock 25
	REG_BG2CNT = BG_8BPP | BG_CBB(0) | BG_SBB(25) | BG_AFF_64x64;

	// Initialize affine registers for bg 2
	const BG_AFFINE bg_aff_default= { 256, 0, 0, 256, 0, 0 };
	REG_BG_AFFINE[2] = bg_aff_default;

	g_bg_affine = bg_aff_default;

	g_cam_pos.x = 64 << 8;
	g_cam_pos.y = 18 << 8;
	g_cam_pos.z = 83 << 8;


	//NOTE: change the z number to change the camera offset used for rotation calcs
	g_pre_rotate_cam_pos.x = 90 << 8;
	g_pre_rotate_cam_pos.y = 18 << 8;
	g_pre_rotate_cam_pos.z = 244 << 8;

	g_ship_position.x = 90 << 8;
	g_ship_position.y = 0;
	g_ship_position.z = 225 << 8;

	g_ship_velocity.x = 0;
	g_ship_velocity.y = 0;
	g_ship_velocity.z = 0;

	g_heading_z.x = 0;
	g_heading_z.y = 0;
	g_heading_z.z = -SHIP_SPEED;

	// Sprite testing stuff, load in tile palette and art to use as --
	// placeholder sprite assets

	image_load_spriteset(spriteTileset);
	image_load_palette256(spriteTileset, PALETTE_OBJ);

	// Wait for vblank then init sprite system
	while(REG_VCOUNT != 160) {}

	init_sprites();
	((struct Sprite*) (MEM_OAM))[0].attr3 = cosf16(0);
	((struct Sprite*) (MEM_OAM))[1].attr3 = -sinf16(0);
	((struct Sprite*) (MEM_OAM))[2].attr3 = sinf16(0);
	((struct Sprite*) (MEM_OAM))[3].attr3 = cosf16(0); 

	s16 x, y;
	x = 55;
	y = 60;

	g_offset.x = 0;
	g_offset.y = 0;
	g_offset.z = 1<<8;

	f16 p1 = s16_to_f16(1);
	s16 angle_offset = 0;
	// end sprite testing setup -------------------------------------

	g_skybox_rot = 0;
	g_cam_rotation_deg = 180;

	// Go! 

	sound_note(NOTE_C, 0, 4, 0);

	for (;;) {

		// vsync
		// TODO: make this ISR based so we don't just thrash the
		// CPU. Ideally we would issue a hlt here and go to sleep
		while(REG_VCOUNT != 160) {}


		if (g_cam_rotation_deg < 0) {
			g_cam_rotation_deg = 359;
		} else if (g_cam_rotation_deg > 359) {
			g_cam_rotation_deg = 0;
		}

		if (input_down(KEY_L)) {
			g_cam_pos.x -= 128;
		}

		if (input_down(KEY_R)) {
			g_cam_pos.x += 128;
		}

		// D-pad input detection

		if (input_down(KEY_UP)) {
			//g_cam_pos.y += 128;
		}

		if (input_down(KEY_DOWN)) {
			//g_cam_pos.y -= 128;
		}

		if (input_down(KEY_LEFT)) {
			angle_offset -= 2;
			g_cam_rotation_deg -= 1;

			g_skybox_rot -= 1;
		}

		if (input_down(KEY_RIGHT)) {
			angle_offset+=2;
			g_cam_rotation_deg += 1;

			g_skybox_rot += 1;
		}

		if (angle_offset > 0) {
			angle_offset--;
		} else if (angle_offset < 0) {
			angle_offset++;
		}

		if (angle_offset > 15) {
			angle_offset = 15;
		}

		update_ship();
		update_camera();
		check_for_collisions();
		update_lap_counter();

		// Recompute camera angles

		g_cam_rotation_sin = sinf16(g_cam_rotation_deg);
		g_cam_rotation_cos = cosf16(g_cam_rotation_deg);


		if (angle_offset < -15) {
			angle_offset = -15;
		}

		// Recompute affine matrix

		//affine_rotscale(&bg_affine, &affine_src);
		//REG_BG_AFFINE[2] = bg_affine;

		sprite_set(1, x, y, 0, ATTR0_8BPP | ATTR0_AFF_DBL | ATTR0_SQUARE, ATTR1_SIZE_64x64, 0);

		// Test

		/*
		sprite_set(
			2, 10, 10, 16,
			ATTR0_8BPP | ATTR0_REG | ATTR0_SQUARE,
			ATTR1_SIZE_16x16,
			0);
		*/
		
		s16 angle = abs_angle(-angle_offset);
		rotate_sprite_group(0, angle);

		REG_BG1HOFS = g_skybox_rot;

	};

	return 0;
}

u32 g_engine_freq = 30;
u32 g_engine_volume = 0;

void update_ship() {

	VECTOR heading = vector_rotate_y(&g_heading_z, -g_cam_rotation_deg);

	//accell if hilding gas
	if (input_down(KEY_A)) {
		g_ship_velocity = vector_add(&g_ship_velocity, &heading);
		
		// Engine color on
		
		pal_obj_mem[SHIP_ENGINE_PAL_INDEX] = SHIP_ENGINE_ON_COLOR;	
		
		g_engine_freq = 350;
		g_engine_volume = 7;
	} else {

		// Engine color off

		pal_obj_mem[SHIP_ENGINE_PAL_INDEX] = SHIP_ENGINE_OFF_COLOR;

		g_engine_freq = 100;
		g_engine_volume = 6;
	}

	g_ship_velocity = vector_mul(&g_ship_velocity, 254);


	// Max Speed
	f32 length = vector_sqr_mag(&g_ship_velocity);
	if (length > ((2 << 8) | 64) ) {
		g_ship_velocity = vector_mul(&g_ship_velocity, 250);
	}

	

	if (g_ship_velocity.x > 0) {
		g_ship_velocity.x-=1;
	} else {
		g_ship_velocity.x+=1;
	}

	if (g_ship_velocity.z > 0) {
		g_ship_velocity.z-=1;
	} else {
		g_ship_velocity.z+=1;
	}

	//position update
	g_ship_position = vector_add(&g_ship_position, &g_ship_velocity);

	g_ship_velocity.y = 0;


	REG_SND1FREQ = SFREQ_RESET | g_engine_freq;
	REG_SND1CNT = SSQR_ENV_BUILD(g_engine_volume, 0, 0) | SSQR_DUTY3_4;

}

void update_camera() {

	//follow ship at same speed
	g_pre_rotate_cam_pos  = vector_add(&g_pre_rotate_cam_pos, &g_ship_velocity);

	//calc final cam pos from unrotated position
	VECTOR rot_org = vector_sub(&g_pre_rotate_cam_pos, &g_ship_position);
	VECTOR r = vector_rotate_y(&rot_org, -g_cam_rotation_deg);
	g_cam_pos = vector_add(&r, &g_ship_position);
}

void check_for_collisions() {
	//extract the id of the tile we are on
	u32 x = f32_to_u32(g_ship_position.x);
	u32 y = f32_to_u32(g_ship_position.z);
	x /= 8;
	y /= 8;

	u32 index = x/4 + y*16;
	u32 byte_index = x%4;

	u32* map_mem = (u32*) se_mem[25];
	u32 map_tile_data = (map_mem[index] & (0xFF << (byte_index*8))) >> (byte_index*8);

	if (map_tile_data > 1 && map_tile_data != 4) {

		//We have a collision, resolve

		VECTOR travel_opposite = vector_negate(&g_ship_velocity);
		g_ship_position = vector_add(&travel_opposite, &g_ship_position);
		g_pre_rotate_cam_pos = vector_add(&travel_opposite, &g_pre_rotate_cam_pos);
		travel_opposite = vector_mul(&travel_opposite, 128);

		g_ship_velocity = travel_opposite;

		sound_noise(7, 0b010, 3);
		
	}

}

void update_lap_counter() {

	static flag = 0;

	text_write(g_lap_num, 15, 0, 3, 0);

	if (!cross_flag) {
		if (g_ship_position.z > (240 << 8) ) {
			cross_flag = 1;
			if (first_cross_flag) {
				first_cross_flag = 0;
			} else {
				g_lap_num[0]++;
			}
		}
	} else {
		if (g_ship_position.z < (240 << 8) && g_ship_position.x > (256 << 8)) {
			cross_flag = 0;
		}
	}

	if (g_seconds_left != 0) {
		g_seconds_left-= 1;
		char g_seconds_left_str[] = {"100"};
		sprintf(g_seconds_left_str, "%.3d", g_seconds_left/60);
		text_write(g_seconds_left_str, 15, 0, 26, 0);
	} else {
		g_ship_velocity.x = 0;
		g_ship_velocity.z = 0;

		if (flag) {
		text_write("RUN OVER", 15, 0, 12, 2);
			flag = !flag;
		} else {
			text_write("        ", 15, 0, 12, 2);
			flag = !flag;
		}
	}


}
