#include "filesystem.h"

#include <string.h>

// This is the spacing in bytes that the filesystem initialization uses to
// search for the beginning of the filesystem. Therefore, the FILEYSTEM_INFO
// appended to the binary must appear aligned to this many bytes.

#define filesystem_info_t_ALIGNMENT 512

#define MAGIC_UPPER 0xd0c6ed93
#define MAGIC_LOWER 0x9a04ffa3

#define MULTIBOOT_MAX 0x02040000
#define ROM_MAX       0x0a000000

filesystem_info_t* gFilesystemInfo;
filesystem_error_t gFilesystemError;

void filesystem_init(fileystem_search_start_t start) {
  
	gFilesystemInfo = NULL;
	gFilesystemError = ERR_NONE;

	// Search for the magic number, indicating the filesystem_info_t struct
  
	u32* searchMaximum;
	if (start == FILESYSTEM_SEARCH_MULTIBOOT) {
		searchMaximum = (u32*)MULTIBOOT_MAX;
	} else {
		searchMaximum = (u32*)ROM_MAX;
	}

	u32* currentLocation = (u32*)start;
	while (currentLocation < searchMaximum) {
		if (*currentLocation == MAGIC_UPPER && *(currentLocation + 1) == MAGIC_LOWER) {
			gFilesystemInfo = (filesystem_info_t*)currentLocation;
			return;
		}

		currentLocation += filesystem_info_t_ALIGNMENT / sizeof(u32);
	}

	// We didn't find it in the current search area. If we 
	// were searching the lower multiboot area, try the ROM.
	
	if (start == FILESYSTEM_SEARCH_MULTIBOOT) {
		filesystem_init(FILESYSTEM_SEARCH_ROM);
		return;
	}
	
	// We had no success
	
	gFilesystemError = ERR_FILESYSTEM_NOT_FOUND;

}

filesystem_error_t filesystem_get_error() {
	return gFilesystemError;
}

filesystem_file_t* filesystem_get_next_file(filesystem_file_t* file) {

	// filesystem_file_t structs are four byte aligned, so find the next 
	// point after the end of the file that is four byte aligned.

	char* nextFile = (void*)(file + 1);
	nextFile += file->length;
	
	// Pad out to the next 4 byte boundary

	u32 lengthMod = file->length % sizeof(u32);
	if (lengthMod != 0) {
		s32 padding = -lengthMod + sizeof(32);
		nextFile += padding;
	}

	return (filesystem_file_t*) nextFile;
}

filesystem_file_t* filesystem_get_file(const char* filename) {
	
	// Get the address of the first file, which begins immediately 
	// after filesystem_info_t.

	u32 fileIndex = 0;
	filesystem_file_t* file = (filesystem_file_t*) (gFilesystemInfo + 1);
	
	while (fileIndex < gFilesystemInfo->numFiles) {
		if (strncmp(file->name, filename, 31) == 0) {
			
			// We found it

			return file;
		}
		
		// Move to the next file

		file = filesystem_get_next_file(file);
		fileIndex++;
	}

	// File not found

	gFilesystemError = ERR_FILE_NOT_FOUND;
	return NULL;	

}

void* filesystem_get_file_data(filesystem_file_t* file) {
	return (void*)(file + 1);
}
