#pragma once

#include "tonc_types.h"

/**
 * Simple filesystem implementation that expects a 
 * block of packed files at the end of the ROM.
 * 
 * The packing is as follows:
 * One FILESYSTEM_INFO, followed by one or more sequences of (FILESYSTEM_FILE, file data).
 * 
 * FILESYSTEM_FILE always begins aligned on a 32 bit boundary. If the file data
 * for the previous file does not end aligned on a boundary, it is padded out
 * until the next 32 bit boundary.
 *
 *
 * @author Casey
 */

typedef struct filesystem_info_t {
  
	// Magic number identifying this as a filesystem block: d0c6ed939a04ffa3
	u32 identifierUpper;
	u32 identifierLower;

	// Total filesystem length in bytes
	u32 length;

	// Number of files
	u32 numFiles;

} filesystem_info_t;

typedef struct filesystem_file_t {

	// Filename, maximum 31 chars plus /0
	char name[32];

	// Length in bytes of the file
	u32 length;

} filesystem_file_t;

typedef enum {
	FILESYSTEM_SEARCH_MULTIBOOT = 0x02000000,
	FILESYSTEM_SEARCH_ROM       = 0x08000000
} fileystem_search_start_t;

typedef enum {
	ERR_NONE = 0,
	ERR_FILESYSTEM_NOT_FOUND,
	ERR_FILE_NOT_FOUND
} filesystem_error_t;

/**
 * Initialize the filesystem. 
 * Must be called before performing any filesystem operations.
 *
 * @param start - start address to begin looking for the filesystem.
 */
void filesystem_init(fileystem_search_start_t start);

/**
 * Returns the most recent filesystem error, or ERR_NONE if there was none.
 */
filesystem_error_t filesystem_get_error();

/**
 * Search for a file by name in the filesystem
 *
 * @param filename - the name of the file to look for. Maxmimum 31 characters.
 */
filesystem_file_t* filesystem_get_file(const char* filename);

/**
 * Gets the next file in the filesystem after the provided file,
 * or NULL if there is no such file.
 *
 * @param file - the file preceding the desired file
 */
filesystem_file_t* filesystem_get_next_file(filesystem_file_t* file);

/**
 * Get a pointer to the beginning of a file's data
 * 
 * @param file - file to get the data pointer for
 */
void* filesystem_get_file_data(filesystem_file_t* file);
