#pragma once

#include "filesystem.h"
#include "tonc_types.h"

/**
 * Image loading functions
 *
 * @author Casey
 */

typedef enum {
	IMAGE_4BPP = 0,
	IMAGE_8BPP = 1,
	IMAGE_RAW = 2,
	IMAGE_8BPP2D = 3
} image_format_t;

typedef struct image_tileset_t {
	
	// Format of the tileset, either 4bpp or 8bpp
	image_format_t format;
	
	// Length of image data in bytes
	u32 imageLength;

	// Length of palette data in bytes
	u16 paletteLength;
	
	// This is useless padding
	u16 padding;

} image_tileset_t;

/**
 * Parse a file as a tileset.
 *
 * @param file - file to parse as an image
 * @return pointer to the tileset info struct
 */
image_tileset_t* image_tileset_from_file(filesystem_file_t* file);

/**
 * Get a pointer to the tiles array for a 4bpp tileset 
 * @param tileset - the tileset to use
 * @return array of tiles, or NULL if the image was the wrong type
 */
const TILE4* image_get_tiles4bpp(image_tileset_t* tileset);

/**
 * Get a pointer to the tiles array for a 4bpp tileset 
 * @param tileset - the tileset to use
 * @return array of tiles, or NULL if the image was the wrong type
 */
const TILE8* image_get_tiles8bpp(image_tileset_t* tileset);

/**
 * Get a pointer to the tiles array for a 8bpp tileset, in raw form
 * @param tileset - the tileset to use
 * @return array of tiles, or NULL if the image was the wrong type
 */
const u32* image_get_tilesraw(image_tileset_t* tileset);

/**
 * Load a tileset into a specific charblock
 * @param tileset - tileset to load
 */
void image_load_tileset(image_tileset_t* tileset, u32 charblock);

typedef enum {
	PALETTE_BG = 0,
	PALETTE_OBJ = 1
} palette_t;

/**
 * Load a palette into either BG or OBJ memory.
 * This assumes up to a 256 color palette, so it writes at the beginning of palette mem.
 *
 * @param tileset - tileset to load palette from
 * @param palette - what palette region to load into
 */
void image_load_palette256(image_tileset_t* tileset, palette_t palette);

/**
 * Load a tileset into OBJ memory.
 * This is mostly for testing sprites, one will want to load sprites
 * not from tile data but from sprite sheets with variable size and there own palette
 *
 * @param tileset - tileset to load
 * 
 */
void image_load_spriteset(image_tileset_t* tileset);


