#include "interrupt.h"

#include "tonc_memmap.h"

// Allocate one slot per receiver type

fnptr gIRQReceivers[II_MAX];

typedef struct irq_sender_t {

	// Sender register (offset from REG_BASE)
	u16 reg;
	
	// IRQ bit in the sender register
	u16 flag;

} irq_sender_t;

/**
 * Sender tables from: http://www.coranac.com/tonc/text/interrupts.htm
 */

static const irq_sender_t gIRQSenders[] =
{
    { 0x0004, 0x0008 },     // REG_DISPSTAT,    DSTAT_VBL_IRQ
    { 0x0004, 0x0010 },     // REG_DISPSTAT,    DSTAT_VHB_IRQ
    { 0x0004, 0x0020 },     // REG_DISPSTAT,    DSTAT_VCT_IRQ
    { 0x0102, 0x0040 },     // REG_TM0CNT,      TM_IRQ
    { 0x0106, 0x0040 },     // REG_TM1CNT,      TM_IRQ
    { 0x010A, 0x0040 },     // REG_TM2CNT,      TM_IRQ
    { 0x010E, 0x0040 },     // REG_TM3CNT,      TM_IRQ
    { 0x0128, 0x4000 },     // REG_SCCNT_L      BIT(14) // not sure
    { 0x00BA, 0x4000 },     // REG_DMA0CNT_H,   DMA_IRQ >> 16
    { 0x00C6, 0x4000 },     // REG_DMA1CNT_H,   DMA_IRQ >> 16
    { 0x00D2, 0x4000 },     // REG_DMA2CNT_H,   DMA_IRQ >> 16
    { 0x00DE, 0x4000 },     // REG_DMA3CNT_H,   DMA_IRQ >> 16
    { 0x0132, 0x4000 },     // REG_KEYCNT,      KCNT_IRQ
    { 0x0000, 0x0000 },     // cart: none
};

#define CPU_MODE_MASK 0x1f
#define CPU_MODE_SYSTEM 0x1f
#define CPU_MODE_IRQ 0x12
#define CPU_IRQ_DISABLE (1 << 7)

void irq_isr_dispatch_asm(fnptr isr);

IWRAM_CODE void irq_isr_dispatch() {
	
	u32 enabledInterrupts = REG_IE;
	u32 currentInterrupt = enabledInterrupts & REG_IF;
	
	// REG_IF flags are reset by writing a 1 to them
	
	REG_IF = currentInterrupt;
	REG_IFBIOS = currentInterrupt;

	fnptr isr = NULL;
	for (u32 irqIndex = II_VBLANK; irqIndex < II_MAX; ++irqIndex) {
		if ((1 << irqIndex) & currentInterrupt) {
			
			// We have an ISR and this interrupt was fired

			isr = gIRQReceivers[irqIndex];
			break;
		}
	}

	if (isr == NULL) {
		return;
	}

	// Disable interrupts and call the ISR.
	// The ISR will run in system mode, not irq mode.

	REG_IME = 0;	
	irq_isr_dispatch_asm(isr);
	REG_IME = 1;
	
}

void irq_init() {

	REG_IME = 0;
	REG_ISR_MAIN = irq_isr_dispatch;
	REG_IME = 1;

}

void irq_enable(eIRQIndex irqIndex, fnptr isr) {
	if (irqIndex >= II_MAX) {

		// Invalid index
		return;

	}

	u16 prevIme = REG_IME;
	REG_IME = 0;

	gIRQReceivers[irqIndex] = isr;
	
	// Enable the sender and receiver

	const irq_sender_t* sender = &gIRQSenders[irqIndex];
	*(u16*)(REG_BASE + sender->reg) |= sender->flag;

	REG_IE |= (1 << irqIndex);
	REG_IME = prevIme;
	
}

void irq_disable(eIRQIndex irqIndex) {

	u16 prevIme = REG_IME;
	REG_IME = 0;

	const irq_sender_t* sender = &gIRQSenders[irqIndex];
	*(u16*)(REG_BASE + sender->reg) &= ~(sender->flag);

	REG_IE &= ~(1 << irqIndex);	
	REG_IME = prevIme;

}
