#!/usr/bin/env python2.7

from PIL import Image
import struct
import argparse

TILE_SIZE = 8

if __name__ == "__main__":
	parser = argparse.ArgumentParser(description='Convert image into a variety of formats')
	parser.add_argument('input', help='input image filename')
	parser.add_argument('format', help='output format, either 8bpp, 4bpp, or raw')
	parser.add_argument('output', help='output filename')
	parser.add_argument(
		'--no-palette',
		action='store_false',
		dest='include_palette',
		help='don\'t include the palette')
	
	args = parser.parse_args()

	im = Image.open(args.input)
	size = im.size
	
	print "Image %s:" % args.input
	print "  Mode: %s" % im.mode
	print "  Colors: %s" % len(im.getcolors())
	print "  Size: %sx%s" % (size[0], size[1])
	
	if im.mode != "P":
		print "Error: image must be a palette image"
		print exit(1)

	image_format = None
	if args.format == '8bpp':
		image_format = 1
	elif args.format == '4bpp':
		image_format = 0
	elif args.format == 'raw':
		image_format = 2
	elif args.format == '8bpp2d':
		image_format = 3
	else:
		print 'Unknown image format'
		exit(1)
		
	# Convert into the desired format

	input_pixels = im.getdata()
	output_pixels = bytearray()

	# Iterate over and export tiles

	if args.format == 'raw':
		TILE_SIZE = 1

	if args.format == '8bpp2d':
		TILE_SIZE = 256;

	tile_count = 0
		
	current_box = [0, 0, TILE_SIZE, TILE_SIZE];
	while current_box[1] < size[1]:
		tile_im = im.crop(current_box)
		tile_data = tile_im.getdata()

		if args.format == '8bpp' or args.format == 'raw' or args.format == '8bpp2d':
			output_pixels.extend(tile_data)
		elif args.format == '4bpp':

			# From http://problemkaputt.de/gbatek.htm#lcdvramoverview
			# Each byte representing two dots.
			# the lower 4 bits define the coloxr for the left (!) dot,
			# the upper 4 bits the color for the right dot.
		
			for i in range(0, len(tile_data), 2):

				# Left pixel is i, right pixel is i + 1
				
				left = tile_data[i] & 0xf
				right = (tile_data[i + 1] & 0xf) << 4			
				output_pixels.append(left | right)
		
		# Move the box to the next tile
		
		current_box[0] += TILE_SIZE			
		if current_box[0] >= size[0]:
			current_box[0] = 0
			current_box[1] += TILE_SIZE

		current_box[2] = current_box[0] + TILE_SIZE
		current_box[3] = current_box[1] + TILE_SIZE

		tile_count += 1	

	print "Counted %s tiles" % tile_count
	
	# Generate header and optional palette info

	lut = im.resize((len(im.getcolors()), 1))
	lut.putdata(range(len(im.getcolors())))
	palette_colors = list(lut.convert("RGB").getdata())
	
	# Palette index 0 is always transparent, we only export > 0
	# Convert palette colors into 16 bit RGB

	normalized_colors = bytearray()	
	if args.include_palette:
		for color in palette_colors:
			normalized_color = []

			for component in color:
				normalized_color.append(int(round(component / 255.0 * 31)))

			final_color = \
			  normalized_color[0] |\
			  (normalized_color[1] << 5) |\
			  (normalized_color[2] << 10)
			
			normalized_colors.extend(struct.pack('h', final_color))
			
	header = struct.pack(
		'IIHH',
		image_format,
		len(output_pixels),
		len(normalized_colors),
		0)

	print "Wrote %s pixels" % len(output_pixels)
	
	output_bytes = bytearray()
	output_bytes.extend(header)
	output_bytes.extend(output_pixels)

	if args.include_palette:
		output_bytes.extend(normalized_colors)

	output_file = open(args.output, 'w')
	output_file.write(output_bytes)
	output_file.close()

	print "Wrote output %s" % args.output
	
